# -*- coding: utf-8 -*-
import datetime
import logging
import requests
import werkzeug

from dateutil.parser import parse as parse_dt

from odoo import _, api, SUPERUSER_ID
from odoo.addons.saas_worker import metabase
from odoo.addons.saas_worker.compat import http, route, request, RegistryManager
from odoo.addons.saas_worker.const import TIMEOUT, WEBSITE
from odoo.addons.saas_worker.util import from_role, guess_admin_id, replace_session, oauth2_token_to_user, staff_action_authorized
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.exceptions import UserError
from odoo.tools.parse_version import parse_version

try:
    from odoo.addons.base.controllers.rpc import RPC
except ImportError:
    RPC = None


_logger = logging.getLogger(__name__)
class SaaSTrial(http.Controller):

    @route('/saas_trial/upgrade', type='http', auth='user')
    def upgrade(self, **kwargs):
        # Update instance informations
        request.env['publisher_warranty.contract'].sudo().browse().update_notification()

        db_info = metabase.get(request.session.db)

        redirect_url = WEBSITE + '/odoo-enterprise/upgrade-online/%s' % db_info.uuid + '?' + werkzeug.urls.url_encode(kwargs)

        return werkzeug.utils.redirect(redirect_url)

    @route('/saas_trial/database_validated', auth='user')
    def database_validated(self, **kwargs):
        return request.render("saas_trial.database_validated")

    @from_role('accounts')
    @route('/saas_trial/set_oauth_uid', auth='none', type='json', readonly=False)
    def set_oauth_uid(self, login, token, oauth_uid):
        ResPartnerSudo = request.env['res.partner'].sudo()
        partner = ResPartnerSudo._signup_retrieve_partner(token, check_validity=True, raise_exception=True)
        user = request.env['res.users'].sudo().search([
            ('login', '=', login),
            ('partner_id', '=', partner.id)
        ])
        if not user:
            raise UserError(_("Signup token '%s' is not valid") % token)
        user_data = {'oauth_uid': oauth_uid}
        if 'oauth_provider_id' in user._fields:
            user_data['oauth_provider_id'] = False
        # User needeed as hr overrides write to check if the user is HR Officer
        user.with_user(user).sudo().write(user_data)
        request.env.cr.commit()
        # ping odoo.com so that the user can see this db in their /my/databases immediately
        request.env['publisher_warranty.contract'].sudo().with_context(force_ping=True).update_notification(cron_mode=True)
        # return our uuid so odoo.com can easily redirect the user to us via /my/databases/connect/{uuid}
        db = metabase.get(request.env.cr.dbname)
        return {'db_uuid': db.uuid}

    # Invitation gift routes
    @from_role('of')
    @route('/saas_trial/create_invitation', type='json', auth='public')
    def saas_create_invitation(self, login, uid, secret_signature, invitee):
        ResUsers = request.env['res.users'].sudo()
        # >= 12.0 compat
        if uid == SUPERUSER_ID:
            uid = guess_admin_id(request.env.cr)
        user = ResUsers.browse(uid).exists()

        # If the user asking to create invitation isn't found or the signature doesn't match
        if not user or not user._check_secret_signature(secret_signature):
            _logger.error("""saas_trial/create_invitation: secret signature checking failed for mails:
            %s
            Secret received %s for uid %s""", ' '.join([i for i in invitee]), secret_signature, uid)
            return {'result': False}

        if user.login != login:
            _logger.error("""saas_trial/create_invitation: login check failed for mails:\n%s\nlogin %s uid %s""",
                    ' '.join([i for i in invitee]), login, uid)
            return {'result': False}

        return _invite_users([{'mail': k, 'name': v} for k, v in invitee.items()])

    @route('/saas_trial/trial-connect', auth='none', type='http', csrf=False, methods=['POST'])
    def trial_connect(self, token, redirect='/web', **kwargs):
        try:
            oauth_user, impostor_info = oauth2_token_to_user(token)
            if not oauth_user:
                raise werkzeug.exceptions.Forbidden()
        except werkzeug.exceptions.Forbidden:
            url = '/web/login?' + werkzeug.urls.url_encode({'redirect': redirect})
            return werkzeug.utils.redirect(url)

        # avoid altering existing session
        if not request.session.uid:
            if impostor_info:
                if not staff_action_authorized('connect_as', reason=oauth_user.login,
                        impostor_info=impostor_info, impersonated_user_id=oauth_user.id):
                    url = '/web/login?' + werkzeug.urls.url_encode({'redirect': redirect})
                    return werkzeug.utils.redirect(url)

            replace_session(oauth_user.id, impostor_info)
            _logger.info('trial-connect: session injected for user with id=%s', oauth_user.id)

        dbname = request.session.db
        db = metabase.get(dbname)
        if db.mode == 'drop' and db.status == 'unsubscribed' and not db.extra_apps:
            metabase.reset_cron_round(db.name, resettime=True)
            metabase.set_expiry_in(db.name, '3 months', set_mode='trial',
                                   set_status='activated')

        return werkzeug.utils.redirect(redirect)

    @route('/saas_trial/rr/info', type='json', auth='user')
    def rr_info(self, **kwargs):
        # meta info
        info = metabase.get(request.session.db)
        if not info:
            # local testing of customer databases: no meta db!
            return {}

        if not request.env.user.has_group('base.group_user'):
            return {}
        res = metabase.get_rr_status(request.session.db)
        res['is_admin'] = request.env.user.has_group('base.group_erp_manager')

        if res.get('source_version'):
            rr = {parse_version(v): d for v, d in metabase.RR_SCHEDULE.items()}
            pv = parse_version(res['source_version'].replace('-', '~'))

            if pv in rr:
                res['rr_forced_date'] = rr[pv]

                try:
                    if info.rr_locked_until and info.rr_locked_until > parse_dt(res['rr_forced_date']):
                        res['rr_forced_date'] = info.rr_locked_until.strftime("%Y-%m-%d")
                except ValueError:
                    pass


        return res

    @route('/saas_trial/rr/dismiss', type='json', auth='user')
    def rr_dismiss(self, **kwargs):
        # meta info
        info = metabase.get(request.session.db)
        if not info:
            # local testing of customer databases: no meta db!
            return {'requested': False}

        if not request.env.user.has_group('base.group_erp_manager'):
            return {'requested': False}

        return metabase.set_rr_dismissed(request.session.db)

    @route('/saas_trial/rr/production_upgrade', type='json', auth='user')
    def rr_do_upgrade(self, **kwargs):
        # meta info
        info = metabase.get(request.session.db)
        if not info:
            # local testing of customer databases: no meta db!
            return {}

        if not request.env.user.has_group('base.group_erp_manager'):
            return {}

        rr_status = metabase.get_rr_status(request.session.db)
        if rr_status.get('state') in ("rolling_release_available", "rolling_release_warning"):
            db = metabase.get(request.session.db)
            rr_upgrade_versions = "%s -> %s (rr)" % (
                rr_status.get("source_version"),
                rr_status.get("target_version"),
            )
            if not staff_action_authorized('upgrade', reason=rr_upgrade_versions):
                return {
                    "error":
                    "This database is restricted, you need to ask for a staff access authorization "
                    "to the customer. You can do that on the subscription or on the ticket."
                }

            metabase.add_migration(
                db,
                actuator="rr",
                mode="self",
                source_version=rr_status.get("source_version"),
                target_version=rr_status.get("target_version"),
                state_filestore="pending",
            )

            return {'requested': True}
        return {}

    @route('/saas_trial/my_dns', type='http', auth='user')
    def my_dns(self, **kwargs):
        db_info = metabase.get(request.session.db)
        url_params = {
            'action': 'domains',
            'db_name': db_info.name,
            'need_auth': 1,
        }
        url = WEBSITE + '/my/databases?' + werkzeug.urls.url_encode(url_params)

        return werkzeug.utils.redirect(url)

    @route('/saas_trial/substatus', type='json', auth='user')
    def subsription_tatus(self, **kwargs):
        if request.env.user._is_admin() or request.env.user._is_system():
            return {
                'url': WEBSITE + '/my/subscription/status',
                'params': {
                    'db_name': request.env.cr.dbname,
                    'db_uuid': request.env['ir.config_parameter'].sudo().get_param('database.uuid', ''),
                }
            }
        return dict(url='/my', params={})

    @route('/saas_trial/domain_info', type='json', auth='user')
    def domain_info(self, **kwargs):
        db_info = metabase.get(request.session.db)
        data = {'params': {'db_uuid': db_info.uuid}}
        response = requests.post(WEBSITE + '/dns/info', auth=SaasAuth(), json=data, timeout=TIMEOUT).json()
        result = response.get('result', {})

        return result

def _invite_users(invitees, is_multidb=False):
    ResUsers = request.env['res.users'].sudo()

    # retrieve the mails to invite
    invitee_per_mail = {}
    for invitee in invitees:
        invitee_per_mail[invitee['mail']] = invitee
    mails = list(invitee_per_mail)

    create_ctx = dict(
        # In the case of a multidb invitation, the caller should send one email per user.
        # If sent from here, we would instead send of one per user and per database, which is too much.
        no_reset_password=is_multidb,

        # We don't want an update_notification until all properties and groups are written
        no_update_notification=True,
    )

    # in the case of multidb invitations, some users may already exist, and we don't want to create them twice
    if is_multidb:
        existing_mails = set(ResUsers.with_context(**create_ctx).search(['|', ('login', 'in', mails), ('email', 'in', mails)]).mapped('login'))
        new_emails = list(set(mails) - existing_mails)
    else:
        new_emails = mails

    if hasattr(ResUsers, 'web_create_users'): # >= 13.0
        ResUsers.with_context(create_ctx).web_create_users(new_emails)

    invited_users = ResUsers.search([('login', 'in', mails)])

    res = {'result': True, 'users': []}

    # update the name of each user
    # link each user to odoo.com if an oauth_uid is provided
    for u in invited_users:
        res['users'].append({
            'login': u.login,
            'signup_url': u.signup_url,
        })
        vals = {}
        invitee = invitee_per_mail.get(u.login)
        if invitee:
            vals['name'] = invitee['name']
            oauth_uid = invitee.get('oauth_uid')
            if oauth_uid:
                vals['oauth_uid'] = oauth_uid
        else:
            vals['name'] = u.login

        # Writing to users could send a notification, but we will do it after the loop
        u.with_context(no_update_notification=True).write(vals)

    # Now that all properties are written, we can make the update_notification
    request.env['publisher_warranty.contract'].sudo().with_context(force_ping=True).update_notification(cron_mode=True)

    return res


if RPC:
    class OdooRPC(RPC):
        """Handle RPC connections."""

        def _xmlrpc(self, service):
            res = super(OdooRPC, self)._xmlrpc(service)
            # if wrong auth, _xmlrpc will raise, so this code not called
            self.log_external_api_call()
            return res

        @route()
        def jsonrpc(self, service, method, args):
            res = super(OdooRPC, self).jsonrpc(service, method, args)
            self.log_external_api_call()
            return res

        def log_external_api_call(self):
            # Since saas-17.1 the cursor is closed by _check_request in RPC routing layer.
            # A new cursor is opened to update the database.rpc ICP.
            registry = RegistryManager.get(request.session.db)
            with registry.cursor() as cr:
                env = api.Environment(cr, SUPERUSER_ID, {})
                ICP = env['ir.config_parameter'].sudo()
                last = ICP.get_param('database.rpc', '')
                now = str(datetime.date.today())

                if not last or last != now:  # avoid write on each call, once by day is a tradeoff
                    try:
                        cr.execute("""
                            UPDATE ir_config_parameter SET value = %s WHERE id IN (
                                SELECT id from ir_config_parameter WHERE key = 'database.rpc' FOR UPDATE NOWAIT
                            )
                        """, (now,))

                        # if no except due to concurrent update, and no row updated -> try to create it
                        if not cr.rowcount:
                            cr.execute("""
                                INSERT INTO ir_config_parameter (key, value)
                                VALUES('database.rpc', %s)
                            """, (now,))
                    except Exception:
                        pass
